import gql from 'graphql-tag'
import graphqlClient from '../../graphql';
import {uniq} from 'lodash'
import {lookupCloudProviderAlias} from '../../../vue_shared/util.mjs'


class ApplicationBlueprint {

    constructor(source, state) {
        for (const key in source) {
            this[key] = source[key]
        }

        this._state = state
    }


    getDeploymentTemplate(name) {
        return new DeploymentTemplate(this._state['DeploymentTemplate'][name], this._state)
    }

    toJSON() {
        const result = {...this}
        delete result._state
        return result
    }
  
}

class DeploymentTemplate {

    constructor(source, state) {
        for (const key in source) {
            this[key] = source[key]
        }

        this._state = state
    }

    get _primary() {
        return new ResourceTemplate(this._state['ResourceTemplate'][this.primary], this._state)
    }

    toJSON() {
        const result = {...this}
        delete result._state
        return result
    }

}

class ResourceTemplate {
    
    constructor(source, state) {
        for(const key in source) {
            this[key] = source[key]
        }

        this._state = state

        /*
        for(const requirement of this._type.requirements) {
            if(!this.dependencies.find(dep => dep.constraint.name == requirement.name)) {

                this.dependencies.push({
                    __typename: 'Dependency',
                    name: requirement.name,
                    match: null,
                    target: null
                })
            }
        }
        */
    }

    get _type() {
        return this._state['ResourceType'][this.type]
    }

    toJSON() {
        const result = {...this}
        delete result._state
        return result
    }


}


const state = {loaded: false, callbacks: []}
const mutations = {
    setProjectState(state, {key, value}) {
        state[key] = value
    },
    
    loaded(state, status) {
        state.loaded = status
        state.callbacks.forEach(cb => {if(typeof cb == 'function') {cb()} else throw new Error('application blueprint callback is not a function')})  

        state.callbacks = []
    },
    onApplicationBlueprintLoaded(state, cb) { if(state.loaded) {cb()} else state.callbacks.push(cb) }

}
const actions = {
    async fetchProject({commit, dispatch}, params) {
        const {projectPath, fullPath, fetchPolicy, projectGlobal} = params
        commit('loaded', false)
        const query = gql`
          query GetDeploymentTemplateDictionaries($fullPath: ID!) {
              applicationBlueprintProject(fullPath: $fullPath, fetchPolicy: $fetchPolicy) @client {
                  ResourceType
                  ApplicationBlueprint
                  ResourceTemplate
                  DeploymentTemplate
              }
          }
        `

        const result = await graphqlClient.defaultClient.query({
            query,
            variables: {
                ...params,
                fullPath: projectPath || fullPath,
                dehydrated: true,
                fetchPolicy,
            },
            fetchPolicy

        })


        // normalize messy data in here
        const {data, errors} = result
        const root = data.applicationBlueprintProject
        root.projectGlobal = projectGlobal
        if(errors?.length) {
            for(const error of errors) {
                console.error(error)
                throw new Error('Could not fetch project blueprint')
            }
        }
        dispatch('useProjectState', root)


        commit('loaded', true)
    },
    useProjectState({commit, getters}, root) {
        const transforms = {
            ResourceTemplate(resourceTemplate) {
                const generated = getters.getMissingDependencies(resourceTemplate)
                if(generated.length) {
                    for(const generatedDep of generated) {
                        if(!resourceTemplate.dependencies.some(dep => dep.name == generatedDep.name)) {
                            resourceTemplate.dependencies.push(generatedDep)
                        }
                    }
                }
            },
            DeploymentTemplate(deploymentTemplate) {
                if(!deploymentTemplate.resourceTemplates) {
                    deploymentTemplate.resourceTemplates = []
                }
            },
            ApplicationBlueprint(applicationBlueprint) {
                if(!applicationBlueprint.title) {
                    applicationBlueprint.title = applicationBlueprint.name
                }
                applicationBlueprint.projectIcon = root?.projectGlobal?.projectIcon
            },
            ResourceType(resourceType) {
                if(!resourceType.title) {
                    resourceType.title = resourceType.name
                }
            }
        }

        // guarunteed ordering
        const ordering = uniq(['ResourceType', 'ResourceTemplate', 'DeploymentTemplate'].concat(Object.keys(root)))

        for(const key of ordering) {
            const value = root[key]
            if(typeof value != 'object') continue
            if(typeof transforms[key] == 'function')
                Object.values(value).forEach(entry => {if(typeof entry == 'object') {transforms[key](entry)}})
            commit('setProjectState', {key, value})
        }
        if(root.ApplicationBlueprint) {
            commit('setProjectState', {key: 'applicationBlueprint', value: Object.values(root.ApplicationBlueprint)[0]})
        }
    }
    
}

/*
 * TODO try this out
function storeResolver(typename, options) {
    const defaults = {}
    const {instantiateAs} = Object.assign(defaults, options)
    return function(state) {
        const dictionary = state[typename]
        return function(name) {
            let entry
            if(entry = dictionary[name]) {
                if(instantiateAs) {
                    return new instantiateAs(entry, state)
                }
                else return entry
            } else {
                return null
            }
        }
    }
}
*/

const getters = {
    getApplicationRoot(state) {return state},
    resolveResourceType(state) { return name =>  state['ResourceType'][name] },
    resolveResourceTemplate(state) { return name =>  new ResourceTemplate(state['ResourceTemplate'][name], state) },
    resolveDeploymentTemplate(state) { return name =>  new DeploymentTemplate(state['DeploymentTemplate'][name], state) },
    resolveResource(state) { return name =>  state['Resource'][name] },
    //resolveResource: storeResolver('Resource'),
    resolveDeployment(state) { return name =>  state['Deployment'][name] },
    dependenciesFromResourceType(_, getters) {
        return function(resourceTypeName) {
            let resourceType
            if(typeof resourceTypeName == 'string') {
                resourceType = getters.resolveResourceType(resourceTypeName)
            } else { resourceType = resourceTypeName }

            if(!resourceType.requirements) return []
            return resourceType.requirements.map(req => ({
                name: req.name,
                constraint: req,
                match: null,
                target: null,
                __typename: 'Dependency'
            }))
        }
    },
    getMissingDependencies(_, getters) {
        return function(resourceTemplate) {
            const generatedDependencies = getters.dependenciesFromResourceType(resourceTemplate.type)
            if(generatedDependencies.length == resourceTemplate.dependencies?.length) {
                return [] // assuming they're correct
            }
            const result = []
            for(const generated of generatedDependencies) {
                if(!resourceTemplate.dependencies.some(dep => dep.name == generated.name)) {
                    result.push(generated)
                }
            }
            return result
        }
    },
    getApplicationBlueprint(state) { return new ApplicationBlueprint(state.applicationBlueprint, state)},

    /* appease the devtools 
    *  devtools now break when errors occur in getters
    */
    getResources(state) {return Object.values(state.Resource || {})},
    getDeployment(state) {return Object.values(state.Deployment || {})[0]},
    /* appease the devtools */

    applicationBlueprintIsLoaded(state) {return state.loaded},
    getValidResourceTypes(state, getters) {
        return function(dependency, _deploymentTemplate) {
            // having trouble getting fetches finished before the ui starts rendering
            try {
                if(!dependency) return []
                const dependencyName = typeof(dependency) == 'string'? dependency:
                    dependency.name
                //dependency.resourceType || dependency.constraint && dependency.constraint.resourceType.name

                function filteredByType(resourceType) {
                    let typeName = typeof(resourceType) == 'string'? resourceType: resourceType.name
                    return Object.values(state.ResourceType).filter(type => {
                        const isValidImplementation =  Array.isArray(type.extends) && type.extends.includes(typeName)
                        const isConcreteType = Array.isArray(type.implementations) && type.implementations.length > 0
                        return isValidImplementation && isConcreteType
                    })
                }
                let result = filteredByType(dependencyName)

                const deploymentTemplate = typeof _deploymentTemplate == 'string'?
                    getters.resolveDeploymentTemplate(_deploymentTemplate) :
                    new DeploymentTemplate(_deploymentTemplate, state)

                if(result.length == 0 && deploymentTemplate) {
                    const dependency = deploymentTemplate._primary.dependencies
                        .find(dependency => dependency.name == dependencyName)
                    if(dependency) {
                        result = filteredByType(dependency.constraint.resourceType)
                    }
                }

                // TODO query for this information
                const CLOUD_MAPPINGS = {
                    [lookupCloudProviderAlias('gcp')]: 'unfurl.nodes.GoogleCloudObject', 
                    [lookupCloudProviderAlias('aws')]: 'unfurl.nodes.AWSResource',
                    [lookupCloudProviderAlias('azure')]: 'unfurl.nodes.AzureResources', 
                    //[lookupCloudProviderAlias('k8s')]: unknown
                }

                if(deploymentTemplate?.cloud) {
                    const allowedCloudVendor = lookupCloudProviderAlias(deploymentTemplate.cloud)
                    result = result.filter(type => {
                        return !type.extends.includes('unfurl.nodes.CloudObject') ||
                            type.extends.includes(CLOUD_MAPPINGS[allowedCloudVendor])
                    })
                }

                return result
            } catch(e) {
                console.error(e)
                return []
            }
        }
    }
}


export default {state, mutations, actions, getters}
